/*	libprefs.cpp	*/
/*	$Id: libprefs.cpp,v 1.1 1997/06/06 18:37:30 hplus Exp hplus $	*/

/*	include necessary files	*/
#if !defined(LIBPREFS_H)
#include "libprefs.h"
#endif

#if !defined(_OS_H)
#include <OS.h>
#endif

#if !defined(_FIND_DIRECTORY_H)
#include <find_directory.h>
#endif

#if !defined(_MESSAGE_H)
#include <Message.h>	/*	requires C++	*/
#endif

#if !defined(_STRING_H)
#include <string.h>
#endif

#if !defined(_SYS_STAT_H)
#include <sys/stat.h>
#endif

#if !defined(_FCNTL_H)
#include <fcntl.h>
#endif

#if !defined(_UNISTD_H)
#include <unistd.h>
#endif

#if !defined(_DEBUG_H)
#include <Debug.h>
#endif

#if !defined(_STDLIB_H)
#include <stdlib.h>
#endif

#if !defined(_ERRNO_H)
#include <errno.h>
#endif


/*	identifier for this implementation of the prefs library	*/
#define	PREF_VERSION_STRING		"simple libprefs.so version 1.1.0"
/*	this is the interval that pref files are polled at - can do with some tweaking	*/
/*	the environment variable POLL_INTERVAL will override this value	*/
long long	POLL_INTERVAL = 3000000;

/*	some magic system calls	*/
extern "C" int _klock_node_(int fd);
extern "C" int _kunlock_node_(int fd);


#if 0

/*	semaphore debugging code	*/
static status_t my_acquire_sem(sem_id sem, const char * file, int line)
{
	printf("ACQUIRE sem: %d  file: %s  line: %d\n", sem, file, line);
	return acquire_sem(sem);
}

static status_t my_release_sem(sem_id sem, const char * file, int line)
{
	printf("RELEASE sem: %d  file: %s  line: %d\n", sem, file, line);
	return release_sem(sem);
}

#define acquire_sem(x) my_acquire_sem(x, __FILE__, __LINE__)
#define release_sem(x) my_release_sem(x, __FILE__, __LINE__)

#endif


/*	an active application	*/
struct _PREFHandle {
		PREFHandle			next;
		int					ref_count;
		char *				appname;
		PREFData			data_list;
};

/*	a currently open preferences set	*/
struct _PREFData {
		PREFData			next;
		int					ref_count;
		PREFHandle			handle;
		char *				prefs;
		BMessage *			data;
		PREFTransaction		transaction;
		void				(*listen_func)(PREFData, void *);
		void *				listen_cookie;
		thread_id			listen_thread;
		sem_id				listen_sem;
		int					listen_running;
};

struct _PREFTransaction {
		int					fd;
		int					ref_count;
		PREFData			data;
};


/*	some necessary globals	*/
static	sem_id			fPrefSem = -1;	/*	guards the Handle and Data structures/lists	*/
static	PREFHandle		fPrefs = NULL;
static	sem_id			fTransSem = -1;	/*	never held across blocking system calls	*/

/*	The current implementation will store common settings in a folder with		*/
/*	the applications name, with the name "<common>" for the common settings, 		*/
/*	and, literally, "<user>" for the user settings. This is so we don't collide	*/
/*	if find_directory() decides to return the same settings directory for the		*/
/*	common case as for the user case. Declare these names.							*/
static char * common_user_name = "<common>";
static char * user_user_name = "<user>";


/*	right now, find_directory is in the headers, but not in the libraries...	*/
status_t
find_directory(
	directory_which	which,
	char			*returned_path,
	int32			path_length)
{
	char * out = "/system/settings";	/*	we hard-code this folder	*/
	switch (which)
	{
	case B_USER_SETTINGS_DIRECTORY:
	case B_COMMON_SETTINGS_DIRECTORY:
		break;
	default:
		return B_ERROR;
	}
	if (strlen(out) > path_length-1)
		return B_ERROR;
	strcpy(returned_path, out);
	return B_OK;
}


const char *
PREFVersion(
	uint32 *			out_version,
	uint32 *			out_min)
{
	*out_version = PREF_CUR_VERSION;
	*out_min = PREF_MIN_VERSION;
	return PREF_VERSION_STRING;
}


status_t
PREFInit(const char * application, PREFHandle * out_handle)
{
	/*	check for a user preference for polling time	*/
	const char * poll = getenv("POLL_INTERVAL");
	if (poll != NULL) {
		char * temp;
		uint32 pt = strtol(poll, &temp, 10);
		if (pt != 0)
			if (pt < 500000)
				if (pt < 4000)
					POLL_INTERVAL = pt * 1000000;
				else
					POLL_INTERVAL = pt * 1000;
			else
				POLL_INTERVAL = pt;
	}

	status_t err = B_OK;
	*out_handle = NULL;
	if (NULL != strchr(application, '/')) return B_BAD_VALUE;

	/*	initialize library if necessary	*/
	if (fPrefSem == -1)
	{
		fPrefSem = create_sem(0, "prefs sem");
		if (fPrefSem < 0) return fPrefSem;
		fTransSem = create_sem(1, "prefs trans sem");
		if (fTransSem < 0) { delete_sem(fPrefSem); fPrefSem = -1; return fTransSem; }
	}
	else
	/*	get semaphore	*/
		err = acquire_sem(fPrefSem);
	if (err < 0) return err;

	/*	find previous instance	*/
	PREFHandle data = fPrefs;
	while (data != NULL)
	{
		if (!strcmp(data->appname, application))
		{
			data->ref_count++;
			*out_handle = data;
			break;
		}
	}

	/*	create new if not found	*/
	if (data == NULL)
	{
		data = new _PREFHandle;
		data->next = fPrefs;
		data->ref_count = 1;
		data->appname = strdup(application);
		data->data_list = NULL;
		if (!data->appname)
		{
			err = B_NO_MEMORY;
			delete data;
		}
		else
		{
			fPrefs = data;
			*out_handle = data;
		}
	}

	/*	release semaphore and go home	*/
	release_sem(fPrefSem);

	return err;
}


static void
_PREFCloseListen(
	PREFData		data)
{
	if (data->listen_running)
	{
		data->listen_running = 0;
		release_sem(data->listen_sem);
		int32 status = 0;
		wait_for_thread(data->listen_thread, &status);
		delete_sem(data->listen_sem);
	}
}


static void
_PREFDoDisposeSet(
	PREFData *		data)
{

	/*	iterate over handle's item to find self	*/
	for (PREFData * ptr = &((*data)->handle->data_list); *ptr != NULL; ptr = &((*ptr)->next))
	{
		if (*ptr == *data)
		{
			/*	unlink item	*/
			*ptr = (*data)->next;
			break;
		}
	}
	/*	dispose all allocated memory	*/
	free((*data)->prefs);
	delete (*data)->data;
	if ((*data)->transaction)
	{
		close((*data)->transaction->fd);
		delete (*data)->transaction;
	}
	/*	dispose item	*/
	delete *data;
	*data = NULL;
}


status_t
PREFShutdown(PREFHandle handle)
{
	status_t err = PREF_NOT_FOUND;
	PREFHandle *data = &fPrefs;
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	iterate over all active handles to find ourselves	*/
	while (*data != NULL)
	{
		if (*data == handle)
		{
			/*	decrease refcount	*/
			(*data)->ref_count--;
			/*	delete and unlink if we're gone	*/
			if ((*data)->ref_count < 1)
			{
				PREFHandle next = (*data)->next;
				/*	here, we should assert that all loaded items are disposed	*/
				/*	and dispose them if not										*/
				for (PREFData *nxptr = NULL, *ptr = &(*data)->data_list; *ptr != NULL; ptr = nxptr)
				{
					nxptr = &(*ptr)->next;
					_PREFDoDisposeSet(ptr);
				}
				delete *data;
				*data = next;
				err = B_OK;
				break;
			}
		}
		data = &(*data)->next;
	}
	/*	release semaphore and go home	*/
	if (fPrefs == NULL)
	{	/*	delete if there's nothing left	*/
		delete_sem(fTransSem);
		fTransSem = -1;
		sem_id del = fPrefSem;
		fPrefSem = -1;
		delete_sem(del);
	}
	else
		release_sem(fPrefSem);

	return err;
}


status_t
PREFLoadSet(PREFHandle handle, const char * module, bool user_level, PREFData * data)
{
	status_t err = B_OK;
	*data = NULL;
	/*	check for bad data	*/
	if (NULL != strchr(module, '/')) return B_BAD_VALUE;
	/*	find the folder in question	*/
	char path[PATH_MAX];
	err = find_directory(user_level ? B_USER_SETTINGS_DIRECTORY : B_COMMON_SETTINGS_DIRECTORY, path, PATH_MAX);
	if (err < B_OK) return err;
	/*	find the application folder in that folder	*/
	if (strlen(module)+strlen(handle->appname)+strlen(path)+30 > PATH_MAX)
		return PREF_TOO_LARGE;
	strcat(path, "/");
	strcat(path, handle->appname);
	mkdir(path, 0x1ff);		/*	this mkdir is harmless	*/
	strcat(path, "/");		/*	this mkdir is harmless	*/
	strcat(path, module);
	mkdir(path, 0x1ff);
	/*	find the storage file in that folder (based on user or common setting)	*/
	strcat(path, "/");
	if (user_level)
		strcat(path, user_user_name);
	else
		strcat(path, common_user_name);
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	find already loaded item	*/
	for (PREFData ptr = handle->data_list; ptr != NULL; ptr = ptr->next)
	{
		if (!strcmp(path, ptr->prefs))
		{
			*data = ptr;
			ptr->ref_count++;
			release_sem(fPrefSem);
			return B_OK;
		}
	}
	/*	create the settings item	*/
	*data = new _PREFData;
	if (!*data)
	{
		release_sem(fPrefSem);
		return B_NO_MEMORY;
	}
	(*data)->next = handle->data_list;
	(*data)->ref_count = 1;
	(*data)->handle = handle;
	(*data)->prefs = strdup(path);
	(*data)->transaction = NULL;
	(*data)->listen_func = NULL;
	(*data)->listen_cookie = NULL;
	(*data)->listen_thread = 0;
	(*data)->listen_sem = 0;
	(*data)->listen_running = 0;
	if (!(*data)->prefs)
	{	/*	yukky error handling code	*/
		delete *data;
		*data = NULL;
		release_sem(fPrefSem);
		return B_NO_MEMORY;
	}
	(*data)->data = NULL;
	err = PREFBeginTransaction(*data, &(*data)->transaction);
	if (err < B_OK)
	{	/*	yukky error handling code	*/
		free((*data)->prefs);
		delete *data;
		*data = NULL;
		release_sem(fPrefSem);
		return err;
	}

	/*	read message and close, if success	*/
	uint32 size = 0;
	lseek((*data)->transaction->fd, 0, SEEK_SET);
	printf("fd %d  ptr %08x  size %d\n", (*data)->transaction->fd, &size, sizeof(size));
	int rd = read((*data)->transaction->fd, &size, sizeof(size));
	if (rd < 0)
	{
		release_sem(fPrefSem);
		return rd;
	}
	(*data)->data = new BMessage;
	if (!(*data)->data)
	{	/*	yukky error handling code	*/
		PREFEndTransaction((*data)->transaction);
		free((*data)->prefs);
		delete *data;
		*data = NULL;
		release_sem(fPrefSem);
		return B_NO_MEMORY;
	}
	if (rd == sizeof(size))
	{
		char * buffer = (char *)malloc(size);
		if (!buffer)
		{
			err = B_NO_MEMORY;
			goto err_exit;
		}
		rd = read((*data)->transaction->fd, buffer, size);
		if (rd != size)
		{
			if (rd < 0)
			{
				err = rd;
				goto err_exit;
			}
			/*	if truncated, we create a clear record with no settings in it	*/
			free(buffer);
			goto no_pref;
		}
		err = (*data)->data->Unflatten(buffer);
		free(buffer);
		if (err < 0) goto err_exit;

		ASSERT(err >= 0);
		/*	deal with various errors in initializing the data	*/
err_exit:
		if (err < 0)
		{	/*	yukky error handling code	*/
			PREFEndTransaction((*data)->transaction);
			free((*data)->prefs);
			delete (*data)->data;
			delete *data;
			*data = NULL;
			release_sem(fPrefSem);
			return B_NO_MEMORY;
		}
	}
no_pref:
	/*	link it in	*/
	handle->data_list = *data;
	PREFEndTransaction((*data)->transaction);
	/*	release semaphore and go home	*/
	release_sem(fPrefSem);
	return B_OK;
}


status_t
PREFReloadSet(PREFData data)
{
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	enter (nest) transaction	*/
	status_t err = PREFBeginTransaction(data, &data->transaction);
	if (err < B_OK) { return err; }
	int32 size = 0;
	/*	read in preferences data	*/
	lseek(data->transaction->fd, 0, 0);
	int rd = read(data->transaction->fd, &size, sizeof(size));
	printf("reload read %d\n", rd);
	if (rd < 0) { err = rd; goto cleanup; }
	if (rd != sizeof(size)) { err = B_ERROR; goto cleanup; }
	char * buffer = (char *)malloc(size);
	if (buffer == NULL) { err = B_NO_MEMORY; goto cleanup; }
	rd = read(data->transaction->fd, buffer, size);
	printf("reload read %d\n", rd);
	if (rd < 0) { err = rd; goto cleanup_buf; }
	if (rd != size) { err = B_IO_ERROR; goto cleanup_buf; }
	/*	unflatten into current preferences message	*/
	err = data->data->Unflatten(buffer);
	printf("reload unflatten %d\n", err);
cleanup_buf:
	free(buffer);
	if (err > 0) err = 0;
cleanup:
	PREFEndTransaction(data->transaction);
	release_sem(fPrefSem);
	return err;
}


status_t
PREFSaveSet(PREFData data)
{
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	enter transaction	*/
	status_t err = PREFBeginTransaction(data, &data->transaction);
	printf("saveset transaction %d\n", err);
	if (err < B_OK) { release_sem(fPrefSem); return err; }
	/*	write message and end transaction	*/
	uint32 size = data->data->FlattenedSize();
	lseek(data->transaction->fd, 0, 0);
	int wr = write(data->transaction->fd, &size, sizeof(size));
	if (wr != sizeof(size))
	{
		PREFEndTransaction(data->transaction);
		release_sem(fPrefSem);
		return (wr < 0) ? wr : B_IO_ERROR;
	}
	char * buffer = (char *)malloc(size);
	if (!buffer)
	{
		PREFEndTransaction(data->transaction);
		release_sem(fPrefSem);
		return B_NO_MEMORY;
	}
	wr = data->data->Flatten(buffer, size);
	if (wr < B_OK)
	{
		PREFEndTransaction(data->transaction);
		free(buffer);
		release_sem(fPrefSem);
		return B_NO_MEMORY;
	}
	wr = write(data->transaction->fd, buffer, size);
	free(buffer);
	if (wr != size)
	{
		PREFEndTransaction(data->transaction);
		release_sem(fPrefSem);
		return (wr < 0) ? wr : B_IO_ERROR;
	}
	/*	end transaction	*/
	printf("saveset done\n", err);
	PREFEndTransaction(data->transaction);
	/*	release semaphore and go home	*/
	release_sem(fPrefSem);
	return B_OK;
}


status_t
PREFDisposeSet(PREFData * data)
{
	/*	start a transaction on the preference to make sure no-one else is using it	*/
	PREFTransaction trans = NULL;
	/*	do not use data->transaction, becuase that will always immediately return (we want exclusive)	*/
	status_t err = PREFBeginTransaction(*data, &trans);
	if (err < 0) return err;

	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	decrease refcount	*/
	if ((*data)->ref_count-- > 1)
	{
		/*	return if not gone	*/
		release_sem(fPrefSem);
		return B_OK;
	}
	_PREFDoDisposeSet(data);
	/*	release semaphore and go home	*/
	release_sem(fPrefSem);
	return B_OK;
}


status_t
PREFSetData(PREFData data, const char * setting, const void * buffer, size_t size, uint32 type)
{
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	call upon message to set data	*/
	data->data->RemoveName(setting);
	status_t err = data->data->AddData(setting, type, buffer, size);
	/*	release semaphore and go home	*/
	release_sem(fPrefSem);
	return err;
}


status_t
PREFDelete(PREFData data, const char * setting)
{
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	call upon message to delete data	*/
	status_t err = data->data->RemoveName(setting);
	/*	release semaphore and go home	*/
	release_sem(fPrefSem);
	return err;
}


status_t
PREFGetData(PREFData data, const char * setting, const void ** out_data, ssize_t * out_size, uint32 * out_type)
{
	/*	acquire semaphore	*/
	if (acquire_sem(fPrefSem) < B_OK) return PREF_NOT_INITED;
	/*	call upon message to get data	*/
	status_t err = data->data->GetInfo(setting, out_type);
	if (err >= B_OK) err = data->data->FindData(setting, *out_type, out_data, out_size);
	/*	release semaphore and go home	*/
	release_sem(fPrefSem);
	return (err >= 0) ? B_OK : PREF_NOT_FOUND;
}


status_t
PREFBeginTransaction(
	PREFData			data,
	PREFTransaction *	transaction)
{
	ASSERT(transaction != NULL);
	if (transaction == NULL) return B_BAD_VALUE;
	*transaction = NULL;
	/*	acquire transaction semaphore	*/
	if (acquire_sem(fTransSem) < B_OK) return PREF_NOT_INITED;
	printf("BeginTransaction <%d>\n", find_thread(NULL));
	/*	test for nesting	*/
	if (data->transaction && (data->transaction == *transaction))
	{
		printf("nest transaction level %d\n", data->transaction->ref_count);
		data->transaction->ref_count++;
		release_sem(fTransSem);
		return B_OK;
	}
	/*	release semaphore while we block for file access	*/
	release_sem(fTransSem);
	/*	open and lock file	*/
open_again:
	int fd = open(data->prefs, O_RDWR | O_CREAT, 0666);
	printf("open %s returns %d\n", data->prefs, fd);
	if ((fd < 0) && (errno == EINTR)) goto open_again;
	if (fd < 0) printf("errno = %08x\n", errno);
	if ((fd < 0) && (errno != EBUSY)) return fd;
	if ((fd < 0) || (_klock_node_(fd) < B_OK))
	{	/*	it is unfortunate that there is no atomic/blocking way to wait	*/
		/*	for exclusive access to a file - now we are forces to busy-wait	*/
		if (fd >= 0) close(fd);
		printf("can't lock - snoozing\n", fd);
		snooze((rand()&0x7fff)+(find_thread(NULL)&0x7fff)+10000);	//	random snooze time to break ties
		goto open_again;
	}
	/*	acquire transaction semaphore - this will always work,							*/
	/*	because the transaction semaphore is never held across blocking system calls	*/
	acquire_sem(fTransSem);
	/*	create transaction object	*/
	*transaction = new _PREFTransaction;
	(*transaction)->ref_count = 1;
	(*transaction)->data = data;
	(*transaction)->fd = fd;
	data->transaction = *transaction;
	/*	release semaphore and go home	*/
	release_sem(fTransSem);
	return B_OK;
}


status_t
PREFEndTransaction(
	PREFTransaction		transaction)
{
	ASSERT(transaction != NULL);
	if (transaction == NULL) return B_BAD_VALUE;
	ASSERT(transaction->data->transaction == transaction);
	if (acquire_sem(fTransSem) < 0) return PREF_NOT_INITED;
	printf("EndTransaction <%d>\n", find_thread(NULL));
	transaction->ref_count--;
	if (transaction->ref_count == 0)
	{
		printf("closing fd %d (%s)\n", transaction->fd, transaction->data->prefs);
		close(transaction->fd);
		transaction->data->transaction = NULL;
		delete transaction;
	}
	release_sem(fTransSem);
	return B_OK;
}


static status_t
_PREFListenFunc(
	void *				param)
{
	/*	this thread function polls the specified settings item for change in its mod date		*/
	PREFData data = (PREFData)param;
	if (!data->listen_func)
	{
		printf("no listen_func???\n");
		return -1;
	}
	struct stat old_stat, new_stat;
	int err = stat(data->prefs, &old_stat);
	if (err < 0) return err;
	while ((acquire_sem_etc(data->listen_sem, 1, B_TIMEOUT, POLL_INTERVAL) == B_TIMED_OUT) &&
		data->listen_running)
	{
		err = stat(data->prefs, &new_stat);
		if (err) return err;
		/*	if mod date differs from what we saw before, we notify the client	*/
		if (new_stat.st_mtime != old_stat.st_mtime)
		{
			(*data->listen_func)(data, data->listen_cookie);
			stat(data->prefs, &old_stat);	/*	update the date stamp (we may have changed it ourselves)	*/
		}
	}
	return B_OK;
}


status_t
PREFListen(
	PREFData			data,
	void				(*listen_func)(
							PREFData			data,
							void *				cookie),
	void *				cookie)
{
	/*	a client wants to be informed when we detect changes in these preferences in "permanent storage"	*/
	if ((data == NULL) || (listen_func == NULL)) return B_BAD_VALUE;
	if (acquire_sem(fPrefSem)) { return PREF_NOT_INITED; }
	/*	if a poller is already currently running, we tell it to die and wait for it	*/
	_PREFCloseListen(data);

	/*	make sure we're in a transaction on the PREFData in question	*/
	PREFTransaction transaction = NULL;
	status_t err = PREFBeginTransaction(data, &transaction);
	printf("listen trans = %08x\n", err);
	if (err < 0) { release_sem(fPrefSem); return err; }
	err = B_OK;

	data->listen_cookie = cookie;
	if (listen_func == NULL)
	{
		/*	setting the listen_func to NULL means we don't want any more notifications	*/
		data->listen_sem = 0;
		data->listen_func = NULL;
		data->listen_thread = 0;
	}
	else
	{
		/*	start up the polling apparatus	*/
		data->listen_sem = create_sem(0, "prefs listen sem");
		if (data->listen_sem < 0) return data->listen_sem;
		data->listen_thread = spawn_thread(_PREFListenFunc, "pref listen func", B_LOW_PRIORITY, data);
		if (data->listen_thread < 0) { delete_sem(data->listen_sem); return data->listen_thread; }
		data->listen_func = listen_func;
		data->listen_running = 1;
	}
	printf("listen end trans\n", err);
	PREFEndTransaction(transaction);
	if (data->listen_thread > 0) err = resume_thread(data->listen_thread);
	release_sem(fPrefSem);
	return err;
}

